@@ -1,10 +1,10 @@
 import re
 import struct
-from .compat import Mapping
-from datetime import datetime, timedelta
+import sys
+from collections.abc import Mapping
+from datetime import datetime, timedelta, timezone
 from io import BytesIO
 
-from .compat import timezone, range, byte_as_integer, unpack_float16
 from .types import (
     CBORDecodeValueError, CBORDecodeEOF, CBORTag, undefined, break_marker,
     CBORSimpleValue, FrozenDict)
@@ -13,7 +13,7 @@ timestamp_re = re.compile(r'^(\d{4})-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)'
                           r'(?:\.(\d{1,6})\d*)?(?:Z|([+-]\d\d):(\d\d))$')
 
 
-class CBORDecoder(object):
+class CBORDecoder:
     """
     The CBORDecoder class implements a fully featured `CBOR`_ decoder with
     several extensions for handling shared references, big integers, rational
@@ -40,7 +40,7 @@ class CBORDecoder(object):
 
     __slots__ = (
         '_tag_hook', '_object_hook', '_share_index', '_shareables', '_fp_read',
-        '_immutable', '_str_errors')
+        '_immutable', '_str_errors', '_stringref_namespace')
 
     def __init__(self, fp, tag_hook=None, object_hook=None,
                  str_errors='strict'):
@@ -50,6 +50,7 @@ class CBORDecoder(object):
         self.str_errors = str_errors
         self._share_index = None
         self._shareables = []
+        self._stringref_namespace = None
         self._immutable = False
 
     @property
@@ -122,6 +123,23 @@ class CBORDecoder(object):
             self._shareables[self._share_index] = value
         return value
 
+    def _stringref_namespace_add(self, string, length):
+        if self._stringref_namespace is not None:
+            next_index = len(self._stringref_namespace)
+            if next_index < 24:
+                is_referenced = length >= 3
+            elif next_index < 256:
+                is_referenced = length >= 4
+            elif next_index < 65536:
+                is_referenced = length >= 5
+            elif next_index < 4294967296:
+                is_referenced = length >= 7
+            else:
+                is_referenced = length >= 11
+
+            if is_referenced:
+                self._stringref_namespace.append(string)
+
     def read(self, amount):
         """
         Read bytes from the data stream.
@@ -144,7 +162,7 @@ class CBORDecoder(object):
             old_index = self._share_index
             self._share_index = None
         try:
-            initial_byte = byte_as_integer(self.read(1))
+            initial_byte = self.read(1)[0]
             major_type = initial_byte >> 5
             subtype = initial_byte & 31
             decoder = major_decoders[major_type]
@@ -183,7 +201,7 @@ class CBORDecoder(object):
         if subtype < 24:
             return subtype
         elif subtype == 24:
-            return byte_as_integer(self.read(1))
+            return self.read(1)[0]
         elif subtype == 25:
             return struct.unpack('>H', self.read(2))[0]
         elif subtype == 26:
@@ -211,19 +229,26 @@ class CBORDecoder(object):
             # Indefinite length
             buf = []
             while True:
-                initial_byte = byte_as_integer(self.read(1))
+                initial_byte = self.read(1)[0]
                 if initial_byte == 0xff:
                     result = b''.join(buf)
                     break
                 elif initial_byte >> 5 == 2:
                     length = self._decode_length(initial_byte & 0x1f)
+                    if length is None or length > sys.maxsize:
+                        raise CBORDecodeValueError(
+                                'invalid length for indefinite bytestring chunk 0x%x' % length
+                                )
                     value = self.read(length)
                     buf.append(value)
                 else:
                     raise CBORDecodeValueError(
                         "non-bytestring found in indefinite length bytestring")
         else:
+            if length > sys.maxsize:
+                raise CBORDecodeValueError('invalid length for bytestring 0x%x' % length)
             result = self.read(length)
+            self._stringref_namespace_add(result, length)
         return self.set_shareable(result)
 
     def decode_string(self, subtype):
@@ -246,19 +271,25 @@ class CBORDecoder(object):
             # that would happily ignore UTF-8 characters split across chunks.
             buf = []
             while True:
-                initial_byte = byte_as_integer(self.read(1))
+                initial_byte = self.read(1)[0]
                 if initial_byte == 0xff:
                     result = ''.join(buf)
                     break
                 elif initial_byte >> 5 == 3:
                     length = self._decode_length(initial_byte & 0x1f)
+                    if length is None or length > sys.maxsize:
+                        raise CBORDecodeValueError(
+                                'invalid length for indefinite string chunk 0x%x' % length)
                     value = self.read(length).decode('utf-8', self._str_errors)
                     buf.append(value)
                 else:
                     raise CBORDecodeValueError(
                         "non-string found in indefinite length string")
         else:
+            if length > sys.maxsize:
+                raise CBORDecodeValueError('invalid length for string 0x%x' % length)
             result = self.read(length).decode('utf-8', self._str_errors)
+            self._stringref_namespace_add(result, length)
         return self.set_shareable(result)
 
     def decode_array(self, subtype):
@@ -276,6 +307,8 @@ class CBORDecoder(object):
                 else:
                     items.append(value)
         else:
+            if length > sys.maxsize:
+                raise CBORDecodeValueError('invalid length for array 0x%x' % length)
             items = []
             if not self._immutable:
                 self.set_shareable(items)
@@ -336,7 +369,11 @@ class CBORDecoder(object):
             return CBORSimpleValue(subtype)
 
         # Major tag 7
-        return special_decoders[subtype](self)
+        try:
+            return special_decoders[subtype](self)
+        except KeyError as e:
+            raise CBORDecodeValueError(
+                    "Undefined Reserved major type 7 subtype 0x%x" % subtype) from e
 
     #
     # Semantic decoders (major tag 6)
@@ -393,15 +430,35 @@ class CBORDecoder(object):
     def decode_fraction(self):
         # Semantic tag 4
         from decimal import Decimal
-        exp, sig = self._decode()
-        return self.set_shareable(Decimal(sig) * (10 ** Decimal(exp)))
+        try:
+            exp, sig = self._decode()
+        except (TypeError, ValueError) as e:
+            raise CBORDecodeValueError("Incorrect tag 4 payload") from e
+        tmp = Decimal(sig).as_tuple()
+        return self.set_shareable(Decimal((tmp.sign, tmp.digits, exp)))
 
     def decode_bigfloat(self):
         # Semantic tag 5
         from decimal import Decimal
-        exp, sig = self._decode()
+        try:
+            exp, sig = self._decode()
+        except (TypeError, ValueError) as e:
+            raise CBORDecodeValueError("Incorrect tag 5 payload") from e
         return self.set_shareable(Decimal(sig) * (2 ** Decimal(exp)))
 
+    def decode_stringref(self):
+        # Semantic tag 25
+        if self._stringref_namespace is None:
+            raise CBORDecodeValueError('string reference outside of namespace')
+
+        index = self._decode()
+        try:
+            value = self._stringref_namespace[index]
+        except IndexError:
+            raise CBORDecodeValueError('string reference %d not found' % index)
+
+        return value
+
     def decode_shareable(self):
         # Semantic tag 28
         old_index = self._share_index
@@ -445,6 +502,14 @@ class CBORDecoder(object):
         from uuid import UUID
         return self.set_shareable(UUID(bytes=self._decode()))
 
+    def decode_stringref_namespace(self):
+        # Semantic tag 256
+        old_namespace = self._stringref_namespace
+        self._stringref_namespace = []
+        value = self._decode()
+        self._stringref_namespace = old_namespace
+        return value
+
     def decode_set(self):
         # Semantic tag 258
         if self._immutable:
@@ -486,14 +551,11 @@ class CBORDecoder(object):
 
     def decode_simple_value(self):
         # XXX Set shareable?
-        return CBORSimpleValue(byte_as_integer(self.read(1)))
+        return CBORSimpleValue(self.read(1)[0])
 
     def decode_float16(self):
         payload = self.read(2)
-        try:
-            value = struct.unpack('>e', payload)[0]
-        except struct.error:
-            value = unpack_float16(payload)
+        value = struct.unpack('>e', payload)[0]
         return self.set_shareable(value)
 
     def decode_float32(self):
@@ -533,12 +595,14 @@ semantic_decoders = {
     3:     CBORDecoder.decode_negative_bignum,
     4:     CBORDecoder.decode_fraction,
     5:     CBORDecoder.decode_bigfloat,
+    25:    CBORDecoder.decode_stringref,
     28:    CBORDecoder.decode_shareable,
     29:    CBORDecoder.decode_sharedref,
     30:    CBORDecoder.decode_rational,
     35:    CBORDecoder.decode_regexp,
     36:    CBORDecoder.decode_mime,
     37:    CBORDecoder.decode_uuid,
+    256:   CBORDecoder.decode_stringref_namespace,
     258:   CBORDecoder.decode_set,
     260:   CBORDecoder.decode_ipaddress,
     261:   CBORDecoder.decode_ipnetwork,
