@@ -1,5 +1,5 @@
-import sys
 import platform
+import struct
 
 import pytest
 
@@ -10,25 +10,34 @@ import cbor2.decoder
 load_exc = ''
 try:
     import _cbor2
-except ImportError as e:
+except ModuleNotFoundError as e:
     if not str(e).startswith('No module'):
         load_exc = str(e)
     _cbor2 = None
 
-cpython33 = pytest.mark.skipif(
-    platform.python_implementation() != "CPython"
-    or sys.version_info < (3, 3)
-    or _cbor2 is None,
-    reason=(load_exc or "requires CPython 3.3+"),
+cpython = pytest.mark.skipif(
+    platform.python_implementation() != "CPython" or _cbor2 is None,
+    reason=(load_exc or "requires CPython"),
 )
 
 
+@pytest.fixture
+def will_overflow():
+    '''
+    Construct an array/string/bytes length which would cause a memory error
+    on decode. This should be less than sys.maxsize (the max integer index)
+    '''
+    bit_size = struct.calcsize('P') * 8
+    huge_length = 1 << (bit_size - 8)
+    return struct.pack('Q', huge_length)
+
+
 class Module(object):
     # Mock module class
     pass
 
 
-@pytest.fixture(params=[pytest.param("c", marks=cpython33), "python"], scope="session")
+@pytest.fixture(params=[pytest.param("c", marks=cpython), "python"], scope="session")
 def impl(request):
     if request.param == "c":
         return _cbor2
