@@ -1,24 +1,18 @@
-from __future__ import division
-
 import math
 import re
-import sys
 from binascii import unhexlify
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, timezone
 from decimal import Decimal
 from email.message import Message
 from fractions import Fraction
 from io import BytesIO
+from ipaddress import ip_address, ip_network
+import struct
+import sys
 from uuid import UUID
-try:
-    from ipaddress import ip_address, ip_network
-except ImportError:
-    def ip_address(x): pass
-    def ip_network(x, strict): pass
 
 import pytest
 
-from cbor2.compat import timezone
 from cbor2.types import FrozenDict
 
 
@@ -359,31 +353,23 @@ def test_datetime_secfrac_requires_digit(impl):
     with pytest.raises(impl.CBORDecodeError) as excinfo:
         impl.loads(b'\xc0\x78\x1a2018-08-02T07:00:59.+00:00')
     assert isinstance(excinfo.value, ValueError)
-    if sys.version_info < (3,):
-        assert str(excinfo.value) == "invalid datetime string: u'2018-08-02T07:00:59.+00:00'"
-    else:
-        assert str(excinfo.value) == "invalid datetime string: '2018-08-02T07:00:59.+00:00'"
+    assert str(excinfo.value) == "invalid datetime string: '2018-08-02T07:00:59.+00:00'"
+
     with pytest.raises(impl.CBORDecodeError) as excinfo:
         impl.loads(b'\xc0\x78\x152018-08-02T07:00:59.Z')
     assert isinstance(excinfo.value, ValueError)
-    if sys.version_info < (3,):
-        assert str(excinfo.value) == "invalid datetime string: u'2018-08-02T07:00:59.Z'"
-    else:
-        assert str(excinfo.value) == "invalid datetime string: '2018-08-02T07:00:59.Z'"
+    assert str(excinfo.value) == "invalid datetime string: '2018-08-02T07:00:59.Z'"
 
 
 def test_bad_datetime(impl):
     with pytest.raises(impl.CBORDecodeError) as excinfo:
         impl.loads(unhexlify('c06b303030302d3132332d3031'))
     assert isinstance(excinfo.value, ValueError)
-    if sys.version_info < (3,):
-        assert str(excinfo.value) == "invalid datetime string: u'0000-123-01'"
-    else:
-        assert str(excinfo.value) == "invalid datetime string: '0000-123-01'"
+    assert str(excinfo.value) == "invalid datetime string: '0000-123-01'"
 
 
 def test_positive_bignum(impl):
-    # Example from RFC 7049 section 3.4.3.
+    # Example from RFC 8949 section 3.4.3.
     decoded = impl.loads(unhexlify('c249010000000000000000'))
     assert decoded == 18446744073709551616
 
@@ -398,6 +384,11 @@ def test_fraction(impl):
     assert decoded == Decimal('273.15')
 
 
+def test_decimal_precision(impl):
+    decoded = impl.loads(unhexlify('c482384dc252011f1fe37d0c70ff50456ba8b891997b07d6'))
+    assert decoded == Decimal('9.7703426561852468194804075821069770622934E-38')
+
+
 def test_bigfloat(impl):
     decoded = impl.loads(unhexlify('c5822003'))
     assert decoded == Decimal('1.5')
@@ -428,10 +419,9 @@ def test_uuid(impl):
     assert decoded == UUID(hex='5eaffac8b51e480581277fdcc7842faf')
 
 
-@pytest.mark.skipif(sys.version_info < (3, 3), reason="Address decoding requires Py3.3+")
 @pytest.mark.parametrize('payload, expected', [
-    ('d9010444c00a0a01', ip_address('192.10.10.1')),
-    ('d901045020010db885a3000000008a2e03707334', ip_address('2001:db8:85a3::8a2e:370:7334')),
+    ('d9010444c00a0a01', ip_address(u'192.10.10.1')),
+    ('d901045020010db885a3000000008a2e03707334', ip_address(u'2001:db8:85a3::8a2e:370:7334')),
     ('d9010446010203040506', (260, b'\x01\x02\x03\x04\x05\x06')),
 ], ids=[
     'ipv4',
@@ -445,7 +435,6 @@ def test_ipaddress(impl, payload, expected):
     assert impl.loads(payload) == expected
 
 
-@pytest.mark.skipif(sys.version_info < (3, 3), reason="Address decoding requires Py3.3+")
 def test_bad_ipaddress(impl):
     with pytest.raises(impl.CBORDecodeError) as exc:
         impl.loads(unhexlify('d9010443c00a0a'))
@@ -457,11 +446,10 @@ def test_bad_ipaddress(impl):
         assert isinstance(exc, ValueError)
 
 
-@pytest.mark.skipif(sys.version_info < (3, 5), reason="Network decoding requires Py3.5+")
 @pytest.mark.parametrize('payload, expected', [
     ('d90105a144c0a800641818', ip_network('192.168.0.100/24', False)),
     ('d90105a15020010db885a3000000008a2e000000001860',
-     ip_network('2001:db8:85a3:0:0:8a2e::/96', False)),
+     ip_network(u'2001:db8:85a3:0:0:8a2e::/96', False)),
 ], ids=[
     'ipv4',
     'ipv6',
@@ -470,13 +458,10 @@ def test_ipnetwork(impl, payload, expected):
     # XXX The following pytest.skip is only included to work-around a bug in
     # pytest under python 3.3 (which prevents the decorator above from skipping
     # correctly); remove when 3.3 support is dropped
-    if sys.version_info < (3, 5):
-        pytest.skip("Network decoding requires Py3.5+")
     payload = unhexlify(payload)
     assert impl.loads(payload) == expected
 
 
-@pytest.mark.skipif(sys.version_info < (3, 5), reason="Network decoding requires Py3.5+")
 def test_bad_ipnetwork(impl):
     with pytest.raises(impl.CBORDecodeError) as exc:
         impl.loads(unhexlify('d90105a244c0a80064181844c0a800001818'))
@@ -528,6 +513,30 @@ def test_cyclic_map(impl):
     assert decoded == {0: decoded}
 
 
+def test_string_ref(impl):
+    decoded = impl.loads(unhexlify('d9010085656669727374d81900667365636f6e64d81900d81901'))
+    assert isinstance(decoded, list)
+    assert decoded[0] == "first"
+    assert decoded[1] == "first"
+    assert decoded[2] == "second"
+    assert decoded[3] == "first"
+    assert decoded[4] == "second"
+
+
+def test_outside_string_ref_namespace(impl):
+    with pytest.raises(impl.CBORDecodeError) as exc:
+        impl.loads(unhexlify('85656669727374d81900667365636f6e64d81900d81901'))
+        assert str(exc.value).endswith('string reference outside of namespace')
+        assert isinstance(exc, ValueError)
+
+
+def test_invalid_string_ref(impl):
+    with pytest.raises(impl.CBORDecodeError) as exc:
+        impl.loads(unhexlify('d9010086656669727374d81900667365636f6e64d81900d81901d81903'))
+        assert str(exc.value).endswith('string reference 3 not found')
+        assert isinstance(exc, ValueError)
+
+
 @pytest.mark.parametrize('payload, expected', [
     ('d9d9f71903e8', 1000),
     ('d9d9f7c249010000000000000000', 18446744073709551616),
@@ -601,7 +610,6 @@ def test_load_from_file(impl, tmpdir):
     assert obj == [1, 10]
 
 
-@pytest.mark.skipif(sys.version_info < (3, 0), reason="No exception with python 2.7")
 def test_nested_exception(impl):
     with pytest.raises((impl.CBORDecodeError, TypeError)) as exc:
         impl.loads(unhexlify('A1D9177082010201'))
@@ -633,16 +641,79 @@ def test_immutable_keys(impl, payload, expected):
     assert value == expected
 
 
-def test_huge_truncated_array(impl):
-    with pytest.raises(impl.CBORDecodeEOF):
-        impl.loads(unhexlify('9b37388519251ae9ca'))
+# Corrupted or invalid data checks
 
-
-def test_huge_truncated_bytes(impl):
-    with pytest.raises((impl.CBORDecodeEOF, MemoryError)):
-        impl.loads(unhexlify('5b37388519251ae9ca'))
+def test_huge_truncated_array(impl, will_overflow):
+    with pytest.raises(impl.CBORDecodeError):
+        impl.loads(unhexlify('9b') + will_overflow)
 
 
 def test_huge_truncated_string(impl):
-    with pytest.raises((impl.CBORDecodeEOF, MemoryError)):
-        impl.loads(unhexlify('7B37388519251ae9ca'))
+    huge_index = struct.pack('Q', sys.maxsize + 1)
+    with pytest.raises((impl.CBORDecodeError, MemoryError)):
+        impl.loads(unhexlify('7b') + huge_index + unhexlify('70717273'))
+
+
+@pytest.mark.parametrize('dtype_prefix', ['7B', '5b'], ids=['string', 'bytes'])
+def test_huge_truncated_data(impl, dtype_prefix, will_overflow):
+    with pytest.raises((impl.CBORDecodeError, MemoryError)):
+        impl.loads(unhexlify(dtype_prefix) + will_overflow)
+
+
+@pytest.mark.parametrize('tag_dtype', ['7F7B', '5f5B'], ids=['string', 'bytes'])
+def test_huge_truncated_indefinite_data(impl, tag_dtype, will_overflow):
+    huge_index = struct.pack('Q', sys.maxsize + 1)
+    with pytest.raises((impl.CBORDecodeError, MemoryError)):
+        impl.loads(unhexlify(tag_dtype) + huge_index + unhexlify('70717273ff'))
+
+
+@pytest.mark.parametrize('data', ['7f61777f6177ffff', '5f41775f4177ffff'], ids=['string', 'bytes'])
+def test_embedded_indefinite_data(impl, data):
+    with pytest.raises(impl.CBORDecodeValueError):
+        impl.loads(unhexlify(data))
+
+
+@pytest.mark.parametrize('data', ['7f01ff', '5f01ff'], ids=['string', 'bytes'])
+def test_invalid_indefinite_data_item(impl, data):
+    with pytest.raises(impl.CBORDecodeValueError):
+        impl.loads(unhexlify(data))
+
+
+@pytest.mark.parametrize('data', [
+    '7f7bff0000000000000471717272ff', '5f5bff0000000000000471717272ff'
+    ],
+    ids=['string', 'bytes']
+)
+def test_indefinite_overflow(impl, data):
+    with pytest.raises(impl.CBORDecodeValueError):
+        impl.loads(unhexlify(data))
+
+
+def test_invalid_cbor(impl):
+    with pytest.raises(impl.CBORDecodeError):
+        impl.loads(unhexlify(
+            'c788370016b8965bdb2074bff82e5a20e09bec21f8406e86442b87ec3ff245b70a47624dc9cdc6824b2a'
+            '4c52e95ec9d6b0534b71c2b49e4bf9031500cee6869979c297bb5a8b381e98db714108415e5c50db7897'
+            '4c271579b01633a3ef6271be5c225eb2'
+            )
+        )
+
+
+@pytest.mark.parametrize('data, expected', [
+    ('fc', '1c'), ('fd', '1d'), ('fe', '1e')
+    ],
+)
+def test_reserved_special_tags(impl, data, expected):
+    with pytest.raises(impl.CBORDecodeValueError) as exc_info:
+        impl.loads(unhexlify(data))
+    assert exc_info.value.args[0] == "Undefined Reserved major type 7 subtype 0x" + expected
+
+
+@pytest.mark.parametrize('data, expected', [
+    ('c400', '4'), ('c500', '5')
+    ],
+)
+def test_decimal_payload_unpacking(impl, data, expected):
+    with pytest.raises(impl.CBORDecodeValueError) as exc_info:
+        impl.loads(unhexlify(data))
+    assert exc_info.value.args[0] == f"Incorrect tag {expected} payload"
