@@ -1,23 +1,17 @@
 import re
-import sys
 from io import BytesIO
 from binascii import unhexlify
 from collections import OrderedDict
-from datetime import datetime, timedelta, date
+from datetime import datetime, timedelta, date, timezone
 from decimal import Decimal
 from email.mime.text import MIMEText
 from fractions import Fraction
+from ipaddress import ip_address, ip_network
 from uuid import UUID
-try:
-    from ipaddress import ip_address, ip_network
-except ImportError:
-    def ip_address(x): pass
-    def ip_network(x, strict): pass
 
 import pytest
 
 from cbor2 import shareable_encoder
-from cbor2.compat import timezone, pack_float16
 from cbor2.types import FrozenDict
 
 
@@ -300,10 +294,9 @@ def test_uuid(impl):
     assert impl.dumps(UUID(hex='5eaffac8b51e480581277fdcc7842faf')) == expected
 
 
-@pytest.mark.skipif(sys.version_info < (3, 3), reason="Address encoding requires Py3.3+")
 @pytest.mark.parametrize('value, expected', [
-    (ip_address('192.10.10.1'), 'd9010444c00a0a01'),
-    (ip_address('2001:db8:85a3::8a2e:370:7334'), 'd901045020010db885a3000000008a2e03707334'),
+    (ip_address(u'192.10.10.1'), 'd9010444c00a0a01'),
+    (ip_address(u'2001:db8:85a3::8a2e:370:7334'), 'd901045020010db885a3000000008a2e03707334'),
 ], ids=[
     'ipv4',
     'ipv6',
@@ -313,10 +306,9 @@ def test_ipaddress(impl, value, expected):
     assert impl.dumps(value) == expected
 
 
-@pytest.mark.skipif(sys.version_info < (3, 3), reason="Network encoding requires Py3.3+")
 @pytest.mark.parametrize('value, expected', [
-    (ip_network('192.168.0.100/24', False), 'd90105a144c0a800001818'),
-    (ip_network('2001:db8:85a3:0:0:8a2e::/96', False),
+    (ip_network(u'192.168.0.100/24', False), 'd90105a144c0a800001818'),
+    (ip_network(u'2001:db8:85a3:0:0:8a2e::/96', False),
      'd90105a15020010db885a3000000008a2e000000001860'),
 ], ids=[
     'ipv4',
@@ -491,21 +483,51 @@ def test_canonical_set(impl, frozen):
     assert serialized == unhexlify('d9010284616161786179626161')
 
 
-@pytest.mark.skipif(sys.version_info >= (3, 6), reason="Using native struct.pack")
-@pytest.mark.parametrize('value, expected', [
-    (3.5, 'F94300'),
-    (100000.0,  False),
-    (3.8, False),
-    (float.fromhex('0x1.0p-24'), 'f90001'),
-    (float.fromhex('0x1.4p-24'), False),
-    (float.fromhex('0x1.ff8p-63'), False),
-    (1e300, False)
-], ids=['float 16', 'float 32', 'float 64',
-        'float 16 minimum positive subnormal', 'mantissa o/f to 32',
-        'exponent o/f to 32', 'oversize float'])
-def test_float16_encoder(value, expected):
-    if expected:
-        expected = unhexlify(expected)
-        assert pack_float16(value) == expected
-    else:
-        assert not pack_float16(value)
+@pytest.mark.parametrize('value', [
+    u'',
+    u'a',
+    u'abcde',
+    b'\x01\x02\x03\x04',
+    [u'a', u'bb', u'a', u'bb'],
+    [u'a', u'bb', u'ccc', u'dddd', u'a', u'bb'],
+    {u'a': u'm', u'bb': u'nn', u'e': u'm', u'ff': u'nn'},
+    {u'a': u'm', u'bb': u'nn', u'ccc': u'ooo', u'dddd': u'pppp',
+     u'e': u'm', u'ff': u'nn'},
+], ids=['empty string', 'short string', 'long string', 'bytestring',
+        'array of short strings', 'no repeated long strings',
+        'dict with short keys and strings',
+        'dict with no repeated long strings'])
+def test_encode_stringrefs_unchanged(impl, value):
+    expected = impl.dumps(value)
+    if isinstance(value, list) or isinstance(value, dict):
+        expected = b'\xd9\x01\x00' + expected
+    assert impl.dumps(value, string_referencing=True) == expected
+
+
+def test_encode_stringrefs_array(impl):
+    value = [u'aaaa', u'aaaa', u'bbbb', u'aaaa', u'bbbb']
+    equivalent = [u'aaaa', impl.CBORTag(25, 0), u'bbbb', impl.CBORTag(25, 0), impl.CBORTag(25, 1)]
+    assert impl.dumps(value, string_referencing=True) == b'\xd9\x01\x00' + impl.dumps(equivalent)
+
+
+def test_encode_stringrefs_dict(impl):
+    value = {u'aaaa': u'mmmm', u'bbbb': u'bbbb', u'cccc': u'aaaa', u'mmmm': u'aaaa'}
+    expected = unhexlify(
+        'd90100' 'a4'
+        '6461616161' '646d6d6d6d'
+        '6462626262' 'd81902'
+        '6463636363' 'd81900'
+        'd81901' 'd81900'
+    )
+    assert impl.dumps(value, string_referencing=True, canonical=True) == expected
+
+
+@pytest.mark.parametrize('tag', [-1, 2**64, 'f'], ids=['too small', 'too large', 'wrong type'])
+def test_invalid_tag(impl, tag):
+    with pytest.raises(TypeError):
+        impl.dumps(impl.CBORTag(tag, 'value'))
+
+
+def test_largest_tag(impl):
+    expected = unhexlify('dbffffffffffffffff6176')
+    assert impl.dumps(impl.CBORTag(2**64-1, 'v')) == expected
